__version__ = "0.1.18"


# from . import a2a_types, server, store, task_manager, task_modifier, task_queue

# __all__ = ["a2a_types", "task_manager", "task_modifier", "server", "store", "task_queue"]



import os

import httpx
from pydantic import BaseModel

from elkar.api_client.models import (
    CreateTaskInput,
    CreateTaskSubscriberRequest,
    DequeueTaskEventInput,
    EnqueueTaskEventInput,
    GetTaskQueryParams,
    TaskEventResponse,
    TaskResponse,
    UnpaginatedOutput,
    UpdateTaskInput,
    UpsertTaskA2AInput,
)


class ElkarClient:
    def __init__(self, base_url: str, api_key: str | None = None) -> None:
        self.base_url = base_url
        self.api_key = api_key or os.getenv("ELKAR_API_KEY")
        if not self.api_key:
            raise ValueError("API key is not set")

    async def make_request(
        self,
        path: str,
        method: str,
        params: BaseModel | None = None,
        query_params: BaseModel | None = None,
    ) -> httpx.Response:
        url = f"{self.base_url}{path}"
        headers = {
            "Content-Type": "application/json",
            "x-api-key": f"{self.api_key}",
        }
        param_dict = params.model_dump(exclude_none=True) if params else None
        query_param_dict = query_params.model_dump(exclude_none=True) if query_params else None

        async with httpx.AsyncClient() as client:
            response = await client.request(method, url, headers=headers, json=param_dict, params=query_param_dict)
        return response

    async def upsert_task(self, params: CreateTaskInput) -> TaskResponse:
        output = await self.make_request("/tasks", "POST", params)
        return TaskResponse.model_validate(output.json())

    async def get_task(self, task_id: str, query_params: GetTaskQueryParams) -> None | TaskResponse:
        output = await self.make_request(
            f"/tasks/{task_id}",
            "GET",
            query_params=query_params,
        )
        if output.status_code == 404:
            return None
        if output.status_code != 200:
            raise Exception(f"Error getting task: {output.status_code} {output.text}")

        return TaskResponse.model_validate(output.json())

    async def update_task(self, task_id: str, params: UpdateTaskInput) -> TaskResponse:
        output = await self.make_request(
            f"/tasks/{task_id}",
            "PUT",
            params,
        )
        return TaskResponse.model_validate(output.json())

    async def enqueue_task_event(self, params: EnqueueTaskEventInput) -> None:
        output = await self.make_request("/task-events/enqueue", "POST", params)
        if output.status_code != 200:
            raise Exception(f"Error enqueuing task event: {output.status_code} {output.text}")
        return None

    async def dequeue_task_event(self, params: DequeueTaskEventInput) -> UnpaginatedOutput:
        output = await self.make_request("/task-events/dequeue", "POST", params)
        if output.status_code != 200:
            raise Exception(f"Error dequeuing task event: {output.status_code} {output.text}")

        return UnpaginatedOutput.model_validate(output.json())

    async def create_task_subscriber(self, params: CreateTaskSubscriberRequest) -> None:
        output = await self.make_request("/task-events/subscribers", "POST", params)
        return None

    async def upsert_task_client_side(self, params: UpsertTaskA2AInput) -> TaskResponse:
        output = await self.make_request("/client-side/tasks", "POST", params)
        return TaskResponse.model_validate(output.json())

    async def get_task_client_side(self, task_id: str) -> TaskResponse:
        output = await self.make_request(f"/client-side/tasks/{task_id}", "GET")
        return TaskResponse.model_validate(output.json())



from abc import abstractmethod
from typing import AsyncIterable, Protocol

from elkar.a2a_types import (
    AgentCard,
    CancelTaskResponse,
    GetTaskPushNotificationResponse,
    GetTaskResponse,
    SendTaskResponse,
    SendTaskStreamingResponse,
    SetTaskPushNotificationResponse,
    TaskIdParams,
    TaskPushNotificationConfig,
    TaskQueryParams,
    TaskSendParams,
)


class A2AClientBase(Protocol):
    @abstractmethod
    async def get_agent_card(self) -> AgentCard:
        """Get the agent card from the server."""
        pass

    @abstractmethod
    async def get_url(self) -> str:
        """Get the URL of the server."""
        pass

    @abstractmethod
    async def get_task(self, task_params: TaskQueryParams) -> GetTaskResponse:
        """Get task details by ID."""
        pass

    @abstractmethod
    async def send_task(self, task_params: TaskSendParams) -> SendTaskResponse:
        """Send a new task to the server."""
        pass

    @abstractmethod
    async def send_task_streaming(self, task_params: TaskSendParams) -> AsyncIterable[SendTaskStreamingResponse]:
        """Send a task with streaming response."""
        pass

    @abstractmethod
    async def cancel_task(self, task_params: TaskIdParams) -> CancelTaskResponse:
        """Cancel a running task."""
        pass

    @abstractmethod
    async def set_task_push_notification(
        self, task_params: TaskPushNotificationConfig
    ) -> SetTaskPushNotificationResponse:
        """Set push notification configuration for a task."""
        pass

    @abstractmethod
    async def get_task_push_notification(self, task_params: TaskIdParams) -> GetTaskPushNotificationResponse:
        """Get push notification configuration for a task."""
        pass

    @abstractmethod
    async def resubscribe_to_task(self, task_params: TaskIdParams) -> AsyncIterable[SendTaskStreamingResponse]:
        """Resubscribe to task events."""
        pass



import asyncio
import logging
from datetime import datetime

from elkar.task_queue.base import TaskEvent, TaskEventManager

logger = logging.getLogger(__name__)


class InMemoryTaskEventQueue(TaskEventManager):
    def __init__(self) -> None:
        self.task_subscribers: dict[tuple[str, str | None], dict[str, asyncio.Queue[TaskEvent]]] = {}
        self.lock = asyncio.Lock()

    async def add_subscriber(
        self,
        task_id: str,
        subscriber_identifier: str,
        is_resubscribe: bool = False,
        caller_id: str | None = None,
    ) -> None:
        if (task_id, caller_id) not in self.task_subscribers:
            if is_resubscribe:
                raise ValueError("Cannot resubscribe to a task that is not subscribed to")
            self.task_subscribers[(task_id, caller_id)] = {}
            self.task_subscribers[(task_id, caller_id)][subscriber_identifier] = asyncio.Queue()

    async def remove_subscriber(self, task_id: str, subscriber_identifier: str, caller_id: str | None = None) -> None:
        if (task_id, caller_id) not in self.task_subscribers:
            raise ValueError("Task not subscribed to")
        if subscriber_identifier not in self.task_subscribers[(task_id, caller_id)]:
            raise ValueError("Caller not subscribed to task")
        del self.task_subscribers[(task_id, caller_id)][subscriber_identifier]

    async def enqueue(self, task_id: str, event: TaskEvent, caller_id: str | None = None) -> None:
        if (task_id, caller_id) not in self.task_subscribers:
            raise ValueError("Task not subscribed to")
        for subscriber in self.task_subscribers[(task_id, caller_id)]:
            await self.task_subscribers[(task_id, caller_id)][subscriber].put(event)

    async def dequeue(
        self,
        task_id: str,
        subscriber_identifier: str,
        caller_id: str | None = None,
    ) -> TaskEvent | None:
        if (task_id, caller_id) not in self.task_subscribers:
            raise ValueError("Task not subscribed to")
        if subscriber_identifier not in self.task_subscribers[(task_id, caller_id)]:
            raise ValueError("Subscriber not subscribed to task")
        queue = self.task_subscribers[(task_id, caller_id)][subscriber_identifier]
        event = await queue.get()
        return event



from abc import abstractmethod
from typing import Protocol

from elkar.a2a_types import Artifact, Message, Task, TaskSendParams, TaskStatus


class TaskModifierBase(Protocol):
    @abstractmethod
    async def get_send_params(self) -> TaskSendParams | None: ...
    @abstractmethod
    async def get_task(self, from_store: bool = False) -> Task: ...

    @abstractmethod
    async def set_status(self, status: TaskStatus, is_final: bool = False) -> None: ...

    @abstractmethod
    async def add_messages_to_history(self, messages: list[Message]) -> None: ...

    @abstractmethod
    async def upsert_artifacts(self, artifacts: list[Artifact]) -> None: ...



from typing import Any

from elkar.json_rpc import JSONRPCError


class TaskNotFoundError(JSONRPCError):
    code: int = -32001
    message: str = "Task not found"
    data: None = None


class TaskNotCancelableError(JSONRPCError):
    code: int = -32002
    message: str = "Task cannot be canceled"
    data: None = None


class PushNotificationNotSupportedError(JSONRPCError):
    code: int = -32003
    message: str = "Push Notification is not supported"
    data: None = None


class UnsupportedOperationError(JSONRPCError):
    code: int = -32004
    message: str = "This operation is not supported"
    data: None = None


class ContentTypeNotSupportedError(JSONRPCError):
    code: int = -32005
    message: str = "Incompatible content types"
    data: None = None


class StreamingNotSupportedError(JSONRPCError):
    code: int = -32006
    message: str = "Streaming is not supported"
    data: None = None


class AuthenticationRequiredError(JSONRPCError):
    code: int = -32007
    message: str = "Authentication required"
    data: None = None


class AuthorizationFailedError(JSONRPCError):
    code: int = -32008
    message: str = "Authorization failed"
    data: None = None


class InvalidTaskStateError(JSONRPCError):
    code: int = -32009
    message: str = "Invalid task state for operation"
    data: None = None


class RateLimitExceededError(JSONRPCError):
    code: int = -32010
    message: str = "Rate limit exceeded"
    data: None = None


class ResourceUnavailableError(JSONRPCError):
    code: int = -32011
    message: str = "Resource unavailable"
    data: None = None


class JSONParseError(JSONRPCError):
    code: int = -32700
    message: str = "Invalid JSON payload"
    data: Any | None = None


class InvalidRequestError(JSONRPCError):
    code: int = -32600
    message: str = "Request payload validation error"
    data: Any | None = None


class MethodNotFoundError(JSONRPCError):
    code: int = -32601
    message: str = "Method not found"
    data: None = None


class InvalidParamsError(JSONRPCError):
    code: int = -32602
    message: str = "Invalid parameters"
    data: Any | None = None


class InternalError(JSONRPCError):
    code: int = -32603
    message: str = "Internal error"
    data: Any | None = None



import asyncio
import logging
from datetime import datetime

from elkar.a2a_types import (
    Artifact,
    Task,
    TaskSendParams,
    TaskState,
    TaskStatus,
)
from elkar.store.base import (
    ClientSideTaskManagerStore,
    StoredTask,
    TaskManagerStore,
    TaskType,
    UpdateTaskParams,
)

logger = logging.getLogger(__name__)


class InMemoryTaskManagerStore(TaskManagerStore):
    def __init__(self) -> None:
        self.tasks: dict[str | None, dict[str, StoredTask]] = {}
        self.lock = asyncio.Lock()

    def caller_tasks(self, caller_id: str | None) -> dict[str, StoredTask] | None:
        return self.tasks.get(caller_id)

    async def upsert_task(
        self,
        params: TaskSendParams,
        is_streaming: bool = False,
        caller_id: str | None = None,
    ) -> StoredTask:
        async with self.lock:
            caller_tasks = self.tasks.get(caller_id)
            if caller_tasks is None:
                self.tasks[caller_id] = {}

            task = self.tasks[caller_id].get(params.id)
            if task is not None:
                if task.caller_id != caller_id:
                    raise ValueError(f"Task {params.id} is already owned by caller {task.caller_id}")
                if task.task.history is None:
                    task.task.history = []

                task.task.history.append(params.message)
                task.updated_at = datetime.now()
                return task
            self.tasks[caller_id][params.id] = StoredTask(
                id=params.id,
                caller_id=caller_id,
                task_type=TaskType.INCOMING,
                is_streaming=is_streaming,
                task=Task(
                    id=params.id,
                    status=TaskStatus(
                        state=TaskState.SUBMITTED,
                        message=params.message,
                        timestamp=datetime.now(),
                    ),
                    sessionId=params.sessionId,
                    history=[params.message],
                    metadata=params.metadata,
                ),
                push_notification=params.pushNotification,
                created_at=datetime.now(),
                updated_at=datetime.now(),
            )
            print(self.tasks)
            return self.tasks[caller_id][params.id]

    async def get_task(
        self,
        task_id: str,
        history_length: int | None = None,
        caller_id: str | None = None,
    ) -> StoredTask | None:
        async with self.lock:
            caller_tasks = self.caller_tasks(caller_id=caller_id)
            if caller_tasks is None:
                return None
            return caller_tasks.get(task_id)

    async def update_task(self, task_id: str, params: UpdateTaskParams) -> StoredTask:
        return await _update_task(self.lock, self.tasks, task_id, params)


async def _update_task(
    lock: asyncio.Lock,
    tasks: dict[str | None, dict[str, StoredTask]],
    task_id: str,
    params: UpdateTaskParams,
) -> StoredTask:
    async with lock:
        if params.caller_id not in tasks:
            raise ValueError("caller id was not found")
        mutable_task = tasks[params.caller_id][task_id].task
        if task_id not in tasks[params.caller_id]:
            raise ValueError(f"Task {task_id} does not exist")

        if params.status is not None:
            mutable_task.status = params.status
            if mutable_task.history is None:
                mutable_task.history = []
            if params.status.message is not None:
                mutable_task.history.append(params.status.message)
        if params.new_messages is not None:
            if mutable_task.history is None:
                mutable_task.history = []
            mutable_task.history.extend(params.new_messages)
        if params.metadata is not None:
            mutable_task.metadata = params.metadata
        if params.artifacts_updates is not None:
            for artifact in params.artifacts_updates:
                await upsert_artifact(mutable_task, artifact)

        if params.push_notification is not None:
            tasks[params.caller_id][task_id].push_notification = params.push_notification
        tasks[params.caller_id][task_id].updated_at = datetime.now()
        return tasks[params.caller_id][task_id]


async def upsert_artifact(task: Task, artifact: Artifact) -> None:
    if task.artifacts is None:
        task.artifacts = []
    for existing_artifact in task.artifacts:
        if existing_artifact.index == artifact.index:
            if existing_artifact.lastChunk == True:
                raise ValueError(f"Artifact {existing_artifact.index} is already a last chunk")
            existing_artifact.parts.extend(artifact.parts)
            existing_artifact.lastChunk = artifact.lastChunk
            return
    task.artifacts.append(artifact)


class InMemoryClientSideTaskManagerStore(ClientSideTaskManagerStore):
    def __init__(self) -> None:
        self.tasks: dict[str | None, dict[str, StoredTask]] = {}
        self.lock = asyncio.Lock()

    async def upsert_task_for_client(self, task: Task, agent_url: str, caller_id: str | None = None) -> StoredTask:
        async with self.lock:
            task_id = task.id
            caller_tasks = self.tasks.get(caller_id)
            if caller_tasks is None:
                caller_tasks = {}
            curr_task = caller_tasks.get(task_id)
            if curr_task is None:
                caller_tasks[task_id] = StoredTask(
                    id=task_id,
                    task=task,
                    task_type=TaskType.OUTGOING,
                    is_streaming=False,
                    push_notification=None,
                    created_at=datetime.now(),
                    updated_at=datetime.now(),
                    caller_id=caller_id,
                    agent_url=agent_url,
                )
                return caller_tasks[task_id]
            if caller_id is not None and curr_task.caller_id != caller_id:
                raise ValueError(f"Task {task_id} is already owned by caller {curr_task.caller_id}")
            elif caller_id is None and curr_task.caller_id is not None:
                raise ValueError(f"Task {task_id} is already owned")
            curr_task.task = task
            curr_task.updated_at = datetime.now()
            curr_task.agent_url = agent_url
            return curr_task

    async def get_task_for_client(self, task_id: str, caller_id: str | None) -> StoredTask | None:
        async with self.lock:
            caller_tasks = self.tasks.get(caller_id)
            if caller_tasks is None:
                return None
            task = caller_tasks.get(task_id)
            if task is not None and task.caller_id == caller_id:
                return task
            return None

    async def update_task(self, task_id: str, params: UpdateTaskParams) -> StoredTask:
        return await _update_task(self.lock, self.tasks, task_id, params)



import json
import logging
from typing import Any, AsyncIterable, Callable

from pydantic import ValidationError
from sse_starlette.sse import EventSourceResponse
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware
from starlette.requests import Request
from starlette.responses import JSONResponse, Response

from elkar.a2a_errors import InternalError, InvalidRequestError, JSONParseError
from elkar.a2a_types import *
from elkar.json_rpc import JSONRPCError
from elkar.task_manager.task_manager_base import RequestContext, TaskManager

logger = logging.getLogger(__name__)


class A2AServer[T: TaskManager]:
    def __init__(
        self,
        task_manager: T,
        host: str = "0.0.0.0",
        port: int = 5000,
        endpoint: str = "/",
        cors_allow_origins: list[str] = ["*"],
        context_extractor: Callable[[Request], RequestContext] | None = None,
    ) -> None:
        self.host = host
        self.port = port
        self.endpoint = endpoint
        self.task_manager = task_manager
        self.cors_allow_origins = cors_allow_origins
        self.context_extractor = context_extractor

        middleware = [
            Middleware(
                CORSMiddleware,
                allow_origins=self.cors_allow_origins,
                allow_methods=["GET", "POST", "OPTIONS"],
                allow_headers=["Content-Type", "Authorization"],
                allow_credentials=True,
            )
        ]

        self.app = Starlette(middleware=middleware)

        self.app.add_route(self.endpoint, self._process_request, methods=["POST", "OPTIONS"])
        self.app.add_route(
            "/.well-known/agent.json",
            self._get_agent_card,
            methods=["GET", "OPTIONS"],
        )

    def start(self, reload_server: bool = False) -> None:
        if self.task_manager is None:
            raise ValueError("request_handler is not defined")

        import uvicorn

        uvicorn.run(
            self.app,
            host=self.host,
            port=self.port,
            reload=reload_server,
        )

    async def extract_request_context(self, request: Request) -> RequestContext:
        """Extracts the context from the request.
        The context is used to identify the caller and the session. Authentication is handled here.

        This method can be overridden by subclasses or replaced by providing a custom
        context_extractor function during initialization.
        """
        if self.context_extractor:
            return self.context_extractor(request)

        return RequestContext(
            caller_id=None,
            metadata={},
        )

    async def _get_agent_card(self, request: Request) -> Response:
        if request.method == "OPTIONS":
            return Response(status_code=200)

        agent_card = await self.task_manager.get_agent_card()
        return JSONResponse(agent_card.model_dump(exclude_none=True))

    async def _process_request(self, request: Request) -> Response | JSONResponse | EventSourceResponse:
        request_context = await self.extract_request_context(request)
        if request.method == "OPTIONS":
            return Response(status_code=200)

        try:
            body = await request.json()
            json_rpc_request = A2ARequest.validate_python(body)
            result: AsyncIterable[Any] | JSONRPCResponse
            if isinstance(json_rpc_request, GetTaskRequest):
                result = await self.task_manager.get_task(json_rpc_request, request_context)
            elif isinstance(json_rpc_request, SendTaskRequest):
                result = await self.task_manager.send_task(json_rpc_request, request_context)
            elif isinstance(json_rpc_request, SendTaskStreamingRequest):
                result = await self.task_manager.send_task_streaming(json_rpc_request, request_context)
            elif isinstance(json_rpc_request, CancelTaskRequest):
                result = await self.task_manager.cancel_task(json_rpc_request, request_context)
            elif isinstance(json_rpc_request, SetTaskPushNotificationRequest):
                result = await self.task_manager.set_task_push_notification(json_rpc_request, request_context)
            elif isinstance(json_rpc_request, GetTaskPushNotificationRequest):
                result = await self.task_manager.get_task_push_notification(json_rpc_request, request_context)
            elif isinstance(json_rpc_request, TaskResubscriptionRequest):
                result = await self.task_manager.resubscribe_to_task(json_rpc_request, request_context)
            else:
                logger.warning(f"Unexpected request type: {type(json_rpc_request)}")
                raise ValueError(f"Unexpected request type: {type(request)}")

            return self._create_response(result)

        except Exception as e:
            raise e
            # return self._handle_exception(e)

    def _handle_exception(self, e: Exception) -> JSONResponse:
        json_rpc_error: JSONRPCError
        if isinstance(e, json.decoder.JSONDecodeError):
            json_rpc_error = JSONParseError()
        elif isinstance(e, ValidationError):
            json_rpc_error = InvalidRequestError(data=json.loads(e.json()))
        else:
            logger.error(f"Unhandled exception: {e}")
            json_rpc_error = InternalError()

        response = JSONRPCResponse(id=None, error=json_rpc_error)
        return JSONResponse(response.model_dump(exclude_none=True), status_code=400)

    def _create_response(self, result: Any) -> JSONResponse | EventSourceResponse:
        if isinstance(result, AsyncIterable):

            async def event_generator(
                result: AsyncIterable[Any],
            ) -> AsyncIterable[dict[str, str]]:
                async for item in result:
                    yield {"data": item.model_dump_json(exclude_none=True)}

            return EventSourceResponse(event_generator(result))
        elif isinstance(result, JSONRPCResponse):
            return JSONResponse(result.model_dump(exclude_none=True))
        else:
            logger.error(f"Unexpected result type: {type(result)}")
            raise ValueError(f"Unexpected result type: {type(result)}")



from abc import abstractmethod
from typing import Protocol

from elkar.a2a_types import *
from elkar.api_client.client import ElkarClient
from elkar.api_client.models import (
    CreateTaskSubscriberRequest,
    DequeueTaskEventInput,
    EnqueueTaskEventInput,
    UnpaginatedOutput,
)
from elkar.json_rpc import JSONRPCError

TaskEvent = TaskStatusUpdateEvent | TaskArtifactUpdateEvent | JSONRPCError


class ElkarClientTaskQueue:
    def __init__(self, base_url: str, api_key: str | None = None) -> None:
        self.elkar_client = ElkarClient(base_url=base_url, api_key=api_key)

    async def add_subscriber(
        self,
        task_id: str,
        subscriber_identifier: str,
        is_resubscribe: bool = False,
        caller_id: str | None = None,
    ) -> None:
        await self.elkar_client.create_task_subscriber(
            CreateTaskSubscriberRequest(
                task_id=task_id,
                subscriber_id=subscriber_identifier,
                caller_id=caller_id,
            )
        )

    async def remove_subscriber(
        self,
        task_id: str,
        subscriber_identifier: str,
        caller_id: str | None = None,
    ) -> None:
        return None  # TODO code the logic in the backend

    async def enqueue(
        self,
        task_id: str,
        event: TaskEvent,
        caller_id: str | None = None,
    ) -> None:
        await self.elkar_client.enqueue_task_event(
            EnqueueTaskEventInput(
                task_id=task_id,
                event=event,
                caller_id=caller_id,
            )
        )
        return None

    async def dequeue(
        self,
        task_id: str,
        subscriber_identifier: str,
        caller_id: str | None = None,
    ) -> TaskEvent | None:
        output = await self.elkar_client.dequeue_task_event(
            DequeueTaskEventInput(
                task_id=task_id,
                subscriber_id=subscriber_identifier,
                limit=1,
            )
        )
        if len(output.records) == 0:
            return None
        return output.records[0].event_data



from __future__ import annotations

from datetime import datetime
from typing import Optional
from uuid import UUID

from pydantic import BaseModel

from elkar.a2a_types import (
    Artifact,
    Message,
    PushNotificationConfig,
    Task,
    TaskPushNotificationConfig,
    TaskSendParams,
    TaskState,
    TaskStatus,
)
from elkar.store.base import TaskType
from elkar.task_queue.base import TaskEvent


class CreateTaskInput(BaseModel):
    counterparty_identifier: Optional[str] = None
    send_task_params: TaskSendParams
    task_type: Optional[TaskType] = None


class TaskResponse(BaseModel):
    a2a_task: Optional[Task] = None
    created_at: datetime
    id: UUID
    push_notification: Optional[TaskPushNotificationConfig] = None
    state: TaskState
    task_type: TaskType

    updated_at: datetime
    counterparty_identifier: Optional[str] = None


class GetTaskQueryParams(BaseModel):
    history_length: Optional[int] = None
    caller_id: Optional[str] = None


class UpdateTaskInput(BaseModel):
    status: Optional[TaskStatus] = None
    artifacts_updates: Optional[list[Artifact]] = None
    new_messages: Optional[list[Message]] = None
    push_notification: Optional[PushNotificationConfig] = None
    caller_id: Optional[str] = None


class EnqueueTaskEventInput(BaseModel):
    task_id: str
    event: TaskEvent
    caller_id: Optional[str] = None


class DequeueTaskEventInput(BaseModel):
    task_id: str
    subscriber_id: str
    limit: Optional[int] = None


class TaskEventResponse(BaseModel):
    id: UUID
    task_id: str
    event_data: TaskEvent


class UnpaginatedOutput(BaseModel):
    records: list[TaskEventResponse]


class CreateTaskSubscriberRequest(BaseModel):
    task_id: str
    subscriber_id: str
    caller_id: Optional[str] = None


class UpsertTaskA2AInput(BaseModel):
    task: Task
    counterparty_identifier: Optional[str] = None
    server_agent_url: str



import asyncio
import logging
import uuid
from abc import abstractmethod
from dataclasses import dataclass
from datetime import datetime
from typing import (
    Any,
    AsyncIterable,
    Awaitable,
    Callable,
    Optional,
)

from elkar.a2a_errors import (
    InternalError,
    InvalidTaskStateError,
    PushNotificationNotSupportedError,
    TaskNotCancelableError,
    TaskNotFoundError,
)
from elkar.a2a_types import (
    AgentCard,
    Artifact,
    CancelTaskRequest,
    CancelTaskResponse,
    GetTaskPushNotificationRequest,
    GetTaskPushNotificationResponse,
    GetTaskRequest,
    GetTaskResponse,
    JSONRPCResponse,
    Message,
    SendTaskRequest,
    SendTaskResponse,
    SendTaskStreamingRequest,
    SendTaskStreamingResponse,
    SetTaskPushNotificationRequest,
    SetTaskPushNotificationResponse,
    Task,
    TaskArtifactUpdateEvent,
    TaskPushNotificationConfig,
    TaskResubscriptionRequest,
    TaskSendParams,
    TaskState,
    TaskStatus,
    TaskStatusUpdateEvent,
    TextPart,
)
from elkar.common import ListTasksRequest, PaginatedResponse
from elkar.json_rpc import JSONRPCError
from elkar.store.base import (
    ListTasksParams,
    StoredTask,
    TaskManagerStore,
    UpdateTaskParams,
)
from elkar.store.in_memory import InMemoryTaskManagerStore
from elkar.task_manager.task_manager_base import RequestContext, TaskManager
from elkar.task_modifier.base import TaskModifierBase
from elkar.task_modifier.task_modifier import TaskModifier
from elkar.task_queue.base import TaskEvent, TaskEventManager
from elkar.task_queue.in_memory import InMemoryTaskEventQueue

logger = logging.getLogger(__name__)


@dataclass
class TaskSendOutput:
    status: TaskStatus | None = None
    new_history_messages: list[Message] | None = None
    new_artifacts: list[Artifact] | None = None
    metadata: dict[str, Any] | None = None


class TaskManagerWithModifier[S: TaskManagerStore, Q: TaskEventManager](TaskManager):
    def __init__(
        self,
        agent_card: AgentCard,
        store: Optional[S] = None,
        queue: Optional[Q] = None,
        send_task_handler: Callable[..., Awaitable[None]] | None = None,
    ):
        self.agent_card = agent_card
        self._send_task_handler = send_task_handler
        self.store: S = store or InMemoryTaskManagerStore()  # type: ignore
        self.queue: Q = queue or InMemoryTaskEventQueue()  # type: ignore

    async def get_agent_card(self) -> AgentCard:
        return self.agent_card

    async def get_task(self, request: GetTaskRequest, request_context: RequestContext | None = None) -> GetTaskResponse:
        stored_task = await self.store.get_task(request.params.id)
        if stored_task is None:
            return GetTaskResponse(
                result=None,
                error=TaskNotFoundError(),
            )
        if request_context is not None and stored_task.caller_id != request_context.caller_id:
            return GetTaskResponse(
                result=None,
                error=JSONRPCError(code=-32003, message="Task not found", data=None),
            )
        return GetTaskResponse(result=stored_task.task)

    async def cancel_task(
        self, request: CancelTaskRequest, request_context: RequestContext | None = None
    ) -> CancelTaskResponse:
        stored_task = await self.store.get_task(request.params.id)
        if stored_task is None:
            return CancelTaskResponse(result=None, error=TaskNotFoundError())

        error = self._check_caller_id(stored_task, request_context)
        if error is not None:
            return CancelTaskResponse(
                result=None,
                error=error,
            )
        if stored_task.task.status.state in [
            TaskState.CANCELED,
            TaskState.FAILED,
            TaskState.COMPLETED,
        ]:
            return CancelTaskResponse(result=None, error=TaskNotCancelableError())
        caller_id = request_context.caller_id if request_context is not None else None
        await self.store.update_task(
            request.params.id,
            UpdateTaskParams(
                status=TaskStatus(
                    state=TaskState.CANCELED,
                    message=None,
                    timestamp=datetime.now(),
                ),
                caller_id=caller_id,
            ),
        )
        return CancelTaskResponse()

    async def _prepare_task_modifier(
        self,
        params: TaskSendParams,
        request_context: RequestContext | None,
        with_queue: bool = False,
        is_streaming: bool = False,
    ) -> TaskModifier[S, Q] | SendTaskResponse:
        if self._send_task_handler is None:
            raise ValueError("send_task_handler is not set")

        stored_task = await self.store.get_task(params.id)

        if stored_task is not None and self._check_caller_id(stored_task, request_context) is not None:
            return SendTaskResponse(
                result=None,
                error=TaskNotFoundError(),
            )
        elif stored_task is not None and (stored_task.task.status.state == TaskState.COMPLETED):
            return SendTaskResponse(
                result=None,
                error=InvalidTaskStateError(),
            )
        elif stored_task is None:
            stored_task = await self.store.upsert_task(
                params,
                caller_id=(request_context.caller_id if request_context is not None else None),
                is_streaming=is_streaming,
            )
        task_modifier: TaskModifier[S, Q] = TaskModifier(
            task=stored_task.task,
            store=self.store,
            queue=self.queue if with_queue else None,
            caller_id=(request_context.caller_id if request_context is not None else None),
        )

        return task_modifier

    async def send_task(
        self, request: SendTaskRequest, request_context: RequestContext | None = None
    ) -> SendTaskResponse:
        if self._send_task_handler is None:
            raise ValueError("send_task_handler is not set")
        params = request.params
        task_modifier = await self._prepare_task_modifier(params, request_context, with_queue=False)
        if isinstance(task_modifier, SendTaskResponse):
            return task_modifier

        try:
            await self._send_task_handler(task_modifier, request_context)
        except Exception as e:
            await task_modifier.set_status(
                TaskStatus(
                    state=TaskState.FAILED,
                    message=Message(role="agent", parts=[TextPart(text="Internal error")]),
                    timestamp=datetime.now(),
                ),
            )
            raise e

        stored_task = await self.store.get_task(params.id)
        if stored_task is None:
            return SendTaskResponse(
                result=None,
                error=TaskNotFoundError(),
            )
        return SendTaskResponse(
            jsonrpc="2.0",
            id=None,
            result=stored_task.task,
            error=None,
        )

    async def _send_task_streaming(
        self,
        task_modifier: TaskModifier[S, Q],
        request_context: RequestContext | None = None,
    ) -> None:
        if self._send_task_handler is None:
            raise ValueError("send_task_handler is not set")
        try:
            await self._send_task_handler(task_modifier, request_context)
        except Exception as e:
            await task_modifier.set_status(
                TaskStatus(
                    state=TaskState.FAILED,
                    message=Message(
                        role="agent",
                        parts=[TextPart(text="Internal error in the task handler.")],
                    ),
                    timestamp=datetime.now(),
                ),
                is_final=True,
            )
            raise e

    async def send_task_streaming(
        self,
        request: SendTaskStreamingRequest,
        request_context: RequestContext | None = None,
    ) -> AsyncIterable[SendTaskStreamingResponse] | JSONRPCResponse:
        subscriber_identifier = str(uuid.uuid4())
        params = request.params
        task_modifier = await self._prepare_task_modifier(params, request_context, with_queue=True)

        await self.queue.add_subscriber(
            request.params.id,
            subscriber_identifier,
            is_resubscribe=False,
            caller_id=(request_context.caller_id if request_context is not None else None),
        )
        if isinstance(task_modifier, SendTaskResponse):
            return task_modifier
        asyncio.create_task(self._send_task_streaming(task_modifier, request_context))
        try:
            events = await self.dequeue_task_events(
                request.id,
                request.params.id,
                subscriber_identifier,
                request_context.caller_id if request_context is not None else None,
            )
            return events
        except Exception as e:
            logger.error(f"Error while sending task streaming: {e}")
            return JSONRPCResponse(
                id=request.id,
                error=InternalError(message="Internal error"),
            )

    async def set_task_push_notification(
        self,
        request: SetTaskPushNotificationRequest,
        request_context: RequestContext | None = None,
    ) -> SetTaskPushNotificationResponse:
        if not self.agent_card.capabilities.pushNotifications:
            return SetTaskPushNotificationResponse(result=None, error=PushNotificationNotSupportedError())
        task_id = request.params.id
        task = await self.store.get_task(task_id)
        if task is None:
            return SetTaskPushNotificationResponse(
                result=None,
                error=TaskNotFoundError(),
            )
        error = self._check_caller_id(task, request_context)
        if error is not None:
            return SetTaskPushNotificationResponse(result=None, error=error)

        stored_task = await self.store.update_task(
            task_id,
            UpdateTaskParams(push_notification=request.params.pushNotificationConfig),
        )
        if stored_task.push_notification is None:
            return SetTaskPushNotificationResponse(
                result=None,
                error=PushNotificationNotSupportedError(),
            )
        return SetTaskPushNotificationResponse(
            result=TaskPushNotificationConfig(
                id=task_id,
                pushNotificationConfig=stored_task.push_notification,
            ),
        )

    async def get_task_push_notification(
        self,
        request: GetTaskPushNotificationRequest,
        request_context: RequestContext | None = None,
    ) -> GetTaskPushNotificationResponse:
        task_id = request.params.id
        task = await self.store.get_task(task_id)
        if task is None:
            return GetTaskPushNotificationResponse(
                result=None,
                error=TaskNotFoundError(),
            )
        if task.push_notification is None:
            return GetTaskPushNotificationResponse(
                result=None,
                error=PushNotificationNotSupportedError(),
            )
        return GetTaskPushNotificationResponse(
            result=TaskPushNotificationConfig(
                id=task_id,
                pushNotificationConfig=task.push_notification,
            ),
        )

    async def resubscribe_to_task(
        self,
        request: TaskResubscriptionRequest,
        request_context: RequestContext | None = None,
    ) -> AsyncIterable[SendTaskStreamingResponse] | JSONRPCResponse:
        task_id_params = request.params
        try:
            subscriber_identifier = str(uuid.uuid4())
            await self.queue.add_subscriber(
                task_id_params.id,
                subscriber_identifier,
                is_resubscribe=True,
                caller_id=(request_context.caller_id if request_context is not None else None),
            )
            return await self.dequeue_task_events(
                request.id,
                task_id_params.id,
                subscriber_identifier,
                request_context.caller_id if request_context is not None else None,
            )
        except Exception as e:
            logger.error(f"Error while reconnecting to SSE stream: {e}")
            return JSONRPCResponse(
                id=request.id,
                error=InternalError(message=f"An error occurred while reconnecting to stream: {e}"),
            )

    @staticmethod
    def _check_caller_id(task: StoredTask, request_context: RequestContext | None) -> TaskNotFoundError | None:
        if request_context is not None:
            if task.caller_id != request_context.caller_id:
                return TaskNotFoundError()
        return None

    async def dequeue_task_events(
        self,
        request_id: int | str | None,
        task_id: str,
        subscriber_identifier: str,
        caller_id: str | None,
    ) -> AsyncIterable[SendTaskStreamingResponse] | JSONRPCResponse:
        try:
            return self.try_dequeue_task_events(request_id, task_id, subscriber_identifier, caller_id)

        except Exception as e:
            return JSONRPCResponse(id=request_id, error=InternalError(message=str(e)))

    async def try_dequeue_task_events(
        self,
        request_id: str | int | None,
        task_id: str,
        subscriber_identifier: str,
        caller_id: str | None,
    ) -> AsyncIterable[SendTaskStreamingResponse]:
        while True:
            event = await self.queue.dequeue(task_id, subscriber_identifier, caller_id)
            if event is None:
                continue
            if isinstance(event, JSONRPCError):
                await self.queue.remove_subscriber(task_id, subscriber_identifier, caller_id)
                yield SendTaskStreamingResponse(
                    jsonrpc="2.0",
                    id=request_id,
                    result=None,
                    error=event,
                )
                break
            if isinstance(event, TaskStatusUpdateEvent):
                yield SendTaskStreamingResponse(
                    jsonrpc="2.0",
                    id=request_id,
                    result=event,
                    error=None,
                )
                if event.final:
                    await self.queue.remove_subscriber(task_id, subscriber_identifier, caller_id)
                    break
            if isinstance(event, TaskArtifactUpdateEvent):
                yield SendTaskStreamingResponse(
                    jsonrpc="2.0",
                    id=request_id,
                    result=event,
                    error=None,
                )



from abc import abstractmethod
from typing import Protocol

from elkar.a2a_types import *
from elkar.json_rpc import JSONRPCError

TaskEvent = TaskStatusUpdateEvent | TaskArtifactUpdateEvent | JSONRPCError


class TaskEventManager(Protocol):
    """
    A manager for task events.
    A task event manager is responsible for sending events to subscribers.
    """

    @abstractmethod
    async def add_subscriber(
        self,
        task_id: str,
        subscriber_identifier: str,
        is_resubscribe: bool = False,
        caller_id: str | None = None,
    ) -> None: ...

    @abstractmethod
    async def remove_subscriber(
        self,
        task_id: str,
        subscriber_identifier: str,
        caller_id: str | None = None,
    ) -> None: ...

    @abstractmethod
    async def enqueue(
        self,
        task_id: str,
        event: TaskEvent,
        caller_id: str | None = None,
    ) -> None: ...

    @abstractmethod
    async def dequeue(
        self,
        task_id: str,
        subscriber_identifier: str,
        caller_id: str | None = None,
    ) -> TaskEvent | None: ...



from typing import Any, Literal
from uuid import uuid4

from pydantic import BaseModel, Field


class JSONRPCMessage(BaseModel):
    jsonrpc: Literal["2.0"] = "2.0"
    id: int | str | None = Field(default_factory=lambda: uuid4().hex)


class JSONRPCRequest(JSONRPCMessage):
    method: str
    params: dict[str, Any] | None = None


class JSONRPCError(BaseModel):
    code: int
    message: str
    data: Any | None = None


class JSONRPCResponse(JSONRPCMessage):
    result: Any | None = None
    error: JSONRPCError | None = None



import logging
from dataclasses import dataclass
from typing import AsyncIterable, Optional

from elkar.a2a_types import (
    AgentCard,
    CancelTaskResponse,
    GetTaskPushNotificationResponse,
    GetTaskRequest,
    GetTaskResponse,
    SendTaskRequest,
    SendTaskResponse,
    SendTaskStreamingResponse,
    SetTaskPushNotificationResponse,
    Task,
    TaskArtifactUpdateEvent,
    TaskIdParams,
    TaskPushNotificationConfig,
    TaskSendParams,
    TaskState,
    TaskStatus,
    TaskStatusUpdateEvent,
)
from elkar.client.a2a_client import A2AClient, A2AClientConfig
from elkar.store.base import ClientSideTaskManagerStore, UpdateTaskParams
from elkar.store.in_memory import (
    InMemoryClientSideTaskManagerStore,
)

logger = logging.getLogger(__name__)


@dataclass
class TaskManagerConfig:
    """Configuration for the task manager."""

    client_config: Optional[A2AClientConfig] = None
    store: Optional[ClientSideTaskManagerStore] = None


class ClientSideTaskManager:
    """Manages tasks for the A2A client."""

    def __init__(self, config: TaskManagerConfig, caller_id: str | None):
        self.config = config
        self._store = config.store or InMemoryClientSideTaskManagerStore()
        self._client = A2AClient(config.client_config or A2AClientConfig(base_url=""))

        self._caller_id = caller_id

    async def get_agent_url(self) -> str:
        """Get the agent URL from the server."""
        return await self._client.get_url()

    async def get_agent_card(self) -> AgentCard:
        """Get the agent card from the server."""
        return await self._client.get_agent_card()

    async def send_task(self, request: SendTaskRequest) -> SendTaskResponse:
        """Send a task to the server."""
        task = await self._store.get_task_for_client(task_id=request.params.id, caller_id=self._caller_id)
        if task is not None and task.task.status.state == TaskState.COMPLETED:
            raise ValueError(f"Task {request.params.id} already completed")
        response = await self._client.send_task(request.params)

        # Send task to server

        if response.result:
            agent_url = await self._client.get_url()
            await self._store.upsert_task_for_client(response.result, agent_url, self._caller_id)

        return SendTaskResponse(
            jsonrpc="2.0",
            id=request.id,
            result=response.result,
            error=response.error,
        )

    async def get_task(self, request: GetTaskRequest) -> GetTaskResponse:
        """Get a task from the server."""
        # If not found locally, request from server
        response = await self._client.get_task(request.params)
        if response.result:
            await self._store.upsert_task_for_client(response.result, await self._client.get_url(), self._caller_id)
        return GetTaskResponse(
            jsonrpc="2.0",
            id=request.id,
            result=response.result,
            error=response.error,
        )

    async def send_task_streaming(self, params: TaskSendParams) -> AsyncIterable[SendTaskStreamingResponse]:
        """Send a task to the server with streaming response."""

        # Create task locally first
        stored_task = await self._store.get_task_for_client(task_id=params.id, caller_id=self._caller_id)
        if stored_task is not None and stored_task.task.status.state == TaskState.COMPLETED:
            raise ValueError("Task is already completed")

        # Send task to server with streaming
        stream = await self._client.send_task_streaming(params)
        task = Task(
            id=params.id,
            status=TaskStatus(
                state=TaskState.SUBMITTED,
                message=params.message,
            ),
        )
        await self._store.upsert_task_for_client(task, await self._client.get_url(), self._caller_id)
        async for response in stream:
            if response.result:
                # Update local task with server response
                if isinstance(response.result, TaskStatusUpdateEvent):
                    await self._store.update_task(
                        task.id,
                        UpdateTaskParams(
                            status=response.result.status,
                            caller_id=self._caller_id,
                        ),
                    )

                elif isinstance(response.result, TaskArtifactUpdateEvent):
                    await self._store.update_task(
                        task.id,
                        UpdateTaskParams(
                            artifacts_updates=[response.result.artifact],
                        ),
                    )

            yield response

    async def set_task_push_notification(self, params: TaskPushNotificationConfig) -> SetTaskPushNotificationResponse:
        raise NotImplementedError()

    async def get_task_push_notification(self, params: TaskIdParams) -> GetTaskPushNotificationResponse:
        """Get push notification configuration for a task."""
        raise NotImplementedError()

    async def cancel_task(self, request: TaskIdParams) -> CancelTaskResponse:
        """Cancel a task."""
        stored_task = await self._store.get_task_for_client(task_id=request.id, caller_id=self._caller_id)
        if not stored_task:
            raise ValueError(f"Task {request.id} not found")

        if stored_task.task.status.state in [
            TaskState.COMPLETED,
            TaskState.FAILED,
            TaskState.CANCELED,
        ]:
            raise ValueError(f"Task {request.id} is already in a terminal state")

        params = TaskIdParams(id=request.id)
        cancel_task_response = await self._client.cancel_task(params)
        if cancel_task_response.result:
            await self._store.upsert_task_for_client(
                cancel_task_response.result,
                await self._client.get_url(),
                self._caller_id,
            )
        return cancel_task_response
